def insert_bible_text(event=None):
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    from biblemate.uba.api import run_uba_api
    from biblemate import config

    buffer = event.app.current_buffer if event is not None else text_field.buffer
    selectedText = buffer.copy_selection().text
    references = BibleVerseParser(False).extractAllReferencesReadable(selectedText)
    bible_text = run_uba_api(f"BIBLE:::{config.default_bible}::{references}")
    buffer.insert_text(format_assistant_content(bible_text))
    get_app().reset()



from agentmake import AGENTMAKE_USER_DIR, readTextFile, writeTextFile
import os, shutil, pprint
from pathlib import Path

CONFIG_FILE = os.path.join(os.path.dirname(os.path.realpath(__file__)), "config.py")
CONFIG_FILE_BACKUP = os.path.join(AGENTMAKE_USER_DIR, "biblemate", "config.py")

# NOTE: When add a config item, update both `default_config` and `write_user_config`

# restore config backup after upgrade
default_config = '''banner_title=""
*agent_mode=False
*prompt_engineering=False
*auto_suggestions=True
*max_steps=50
*lite=False
*web_browser=False
*hide_tools_order=True
*default_bible="NET"
*default_commentary="CBSC"
*default_encyclopedia="ISB"
*default_lexicon="Morphology"
*max_semantic_matches=15
*max_log_lines=2000
*mcp_port=33333
*embedding_model="paraphrase-multilingual"
*disabled_tools=['search_1_chronicles_only',
'search_1_corinthians_only',
'search_1_john_only',
'search_1_kings_only',
'search_1_peter_only',
'search_1_samuel_only',
'search_1_thessalonians_only',
'search_1_timothy_only',
'search_2_chronicles_only',
'search_2_corinthians_only',
'search_2_john_only',
'search_2_kings_only',
'search_2_peter_only',
'search_2_samuel_only',
'search_2_thessalonians_only',
'search_2_timothy_only',
'search_3_john_only',
'search_acts_only',
'search_amos_only',
'search_colossians_only',
'search_daniel_only',
'search_deuteronomy_only',
'search_ecclesiastes_only',
'search_ephesians_only',
'search_esther_only',
'search_exodus_only',
'search_ezekiel_only',
'search_ezra_only',
'search_galatians_only',
'search_genesis_only',
'search_habakkuk_only',
'search_haggai_only',
'search_hebrews_only',
'search_hosea_only',
'search_isaiah_only',
'search_james_only',
'search_jeremiah_only',
'search_job_only',
'search_joel_only',
'search_john_only',
'search_jonah_only',
'search_joshua_only',
'search_jude_only',
'search_judges_only',
'search_lamentations_only',
'search_leviticus_only',
'search_luke_only',
'search_malachi_only',
'search_mark_only',
'search_matthew_only',
'search_micah_only',
'search_nahum_only',
'search_nehemiah_only',
'search_numbers_only',
'search_obadiah_only',
'search_philemon_only',
'search_philippians_only',
'search_proverbs_only',
'search_psalms_only',
'search_revelation_only',
'search_romans_only',
'search_ruth_only',
'search_song_of_songs_only',
'search_titus_only',
'search_zechariah_only',
'search_zephaniah_only']'''

if readTextFile(CONFIG_FILE).strip() == "":
    just_upgraded = True
    if os.path.isfile(CONFIG_FILE_BACKUP):
        shutil.copy(CONFIG_FILE_BACKUP, CONFIG_FILE)
    else:
        writeTextFile(CONFIG_FILE, default_config.replace("\n*", "\n"))
else:
    just_upgraded = False

from biblemate import config

def write_user_config(backup=False):
    """Writes the current configuration to the user's config file."""
    configurations = f"""banner_title="{config.banner_title}"
agent_mode={config.agent_mode}
prompt_engineering={config.prompt_engineering}
auto_suggestions={config.auto_suggestions}
max_steps={config.max_steps}
lite={config.lite}
web_browser={config.web_browser}
hide_tools_order={config.hide_tools_order}
default_bible="{config.default_bible}"
default_commentary="{config.default_commentary}"
default_encyclopedia="{config.default_encyclopedia}"
default_lexicon="{config.default_lexicon}"
max_semantic_matches={config.max_semantic_matches}
max_log_lines={config.max_log_lines}
mcp_port={config.mcp_port}
embedding_model="{config.embedding_model}"
disabled_tools={pprint.pformat(config.disabled_tools)}"""
    writeTextFile(CONFIG_FILE_BACKUP if backup else CONFIG_FILE, configurations)

if just_upgraded:
    changed = False
    for config_item in default_config.split("\n*"):
        key, value = config_item.split("=", 1)
        if not hasattr(config, key):
            exec(f"config.{config_item}", globals())
            changed = True
    if changed:
        write_user_config()

# temporary config
config.current_prompt = ""
config.last_book = 43
config.last_chapter = 3
config.last_verse = 16

BIBLEMATE_VERSION = readTextFile(os.path.join(os.path.dirname(os.path.realpath(__file__)), "version.txt"))

# copy etextedit plugins
ETEXTEDIT_USER_PULGIN_DIR = os.path.join(os.path.expanduser("~"), "etextedit", "plugins")
if not os.path.isdir(ETEXTEDIT_USER_PULGIN_DIR):
    Path(ETEXTEDIT_USER_PULGIN_DIR).mkdir(parents=True, exist_ok=True)
BIBLEMATE_ETEXTEDIT_PLUGINS = os.path.join(os.path.dirname(os.path.realpath(__file__)), "etextedit", "plugins")
for file_name in os.listdir(BIBLEMATE_ETEXTEDIT_PLUGINS):
    full_file_name = os.path.join(BIBLEMATE_ETEXTEDIT_PLUGINS, file_name)
    if file_name.endswith(".py") and os.path.isfile(full_file_name) and not os.path.isfile(os.path.join(ETEXTEDIT_USER_PULGIN_DIR, file_name)):
        shutil.copy(full_file_name, ETEXTEDIT_USER_PULGIN_DIR)

AGENTMAKE_CONFIG = {
    "print_on_terminal": False,
    "word_wrap": False,
}
OLLAMA_NOT_FOUND = "`Ollama` is not found! BibleMate AI uses `Ollama` to generate embeddings for semantic searches. You may install it from https://ollama.com/ so that you can perform semantic searches of the Bible with BibleMate AI."
BIBLEMATEDATA = os.path.join(AGENTMAKE_USER_DIR, "biblemate", "data")
if not os.path.isdir(BIBLEMATEDATA):
    Path(BIBLEMATEDATA).mkdir(parents=True, exist_ok=True)

def fix_string(content):
    return content.replace(" ", " ").replace("‑", "-")


from biblemate import config
from biblemate.ui.selection_dialog import TerminalModeDialogs
from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
import re

DIALOGS = TerminalModeDialogs()

# shared dialogs

async def get_multiple_bibles(options, descriptions):
    select = await DIALOGS.getMultipleSelection(
        default_values=[config.default_bible],
        options=options,
        descriptions=descriptions,
        title="Bibles",
        text="Select versions to continue:"
    )
    return select if select else []

async def get_reference(verse_reference=True, exhaustiveReferences=False):
    abbr = BibleBooks.abbrev["eng"]
    input_suggestions = []
    for book in range(1,67):
        input_suggestions += list(abbr[str(book)])
    result = await DIALOGS.getInputDialog(title="Bible Verse Reference", text="Enter a verse reference, e.g. John 3:16", suggestions=input_suggestions)
    if result:
        parser = BibleVerseParser(False)
        result = parser.extractExhaustiveReferencesReadable(result) if exhaustiveReferences else parser.extractAllReferencesReadable(result)
        if result and not verse_reference:
            result = re.sub(r":[\-0-9]+?;", ";", f"{result};")[:-1]
    if result:
        return result
    if not result:
        abbr = BibleBooks.abbrev["eng"]
        book = await DIALOGS.getValidOptions(
            default=str(config.last_book),
            options=[str(book) for book in range(1,67)],
            descriptions=[abbr[str(book)][-1] for book in range(1,67)],
            title="Bible Book",
            text="Select a book to continue:"
        )
        if not book:
            return ""
        config.last_book = book = int(book)
        chapter = await DIALOGS.getValidOptions(
            default=str(config.last_chapter),
            options=[str(chapter) for chapter in range(1,BibleBooks.chapters[int(book)]+1)],
            title="Bible Chapter",
            text="Select a chapter to continue:"
        )
        if not chapter:
            return ""
        config.last_chapter = chapter = int(chapter)
        if verse_reference:
            verse = await DIALOGS.getValidOptions(
                default=str(config.last_verse),
                options=[str(verse) for verse in range(1,BibleBooks.verses[int(book)][int(chapter)]+1)],
                title="Bible Verse",
                text="Select a verse to continue:"
            )
            if not verse:
                return ""
            config.last_verse = verse = int(verse)
            return f"{abbr[str(book)][0]} {chapter}:{verse}"
        return f"{abbr[str(book)][0]} {chapter}"

# dialogs for content retrieval

async def uba_search_bible(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_bible,
        options=options,
        descriptions=descriptions,
        title="Search Bible",
        text="Select a bible version to continue:"
    )
    if not select:
        return ""
    abbr = BibleBooks.abbrev["eng"]
    book = await DIALOGS.getValidOptions(
        default=str(config.last_book),
        options=["0"]+[str(book) for book in range(1,67)],
        descriptions=["ALL"]+[abbr[str(book)][-1] for book in range(1,67)],
        title="Search in Book(s)",
        text="Select all books or a book to continue:"
    )
    if not book:
        return ""
    templates = [
        "search",
        "genesis",
        "exodus",
        "leviticus",
        "numbers",
        "deuteronomy",
        "joshua",
        "judges",
        "ruth",
        "samuel1",
        "samuel2",
        "kings1",
        "kings2",
        "chronicles1",
        "chronicles2",
        "ezra",
        "nehemiah",
        "esther",
        "job",
        "psalms",
        "proverbs",
        "ecclesiastes",
        "songs",
        "isaiah",
        "jeremiah",
        "lamentations",
        "ezekiel",
        "daniel",
        "hosea",
        "joel",
        "amos",
        "obadiah",
        "jonah",
        "micah",
        "nahum",
        "habakkuk",
        "zephaniah",
        "haggai",
        "zechariah",
        "malachi",
        "matthew",
        "mark",
        "luke",
        "john",
        "acts",
        "romans",
        "corinthians1",
        "corinthians2",
        "galatians",
        "ephesians",
        "philippians",
        "colossians",
        "thessalonians1",
        "thessalonians2",
        "timothy1",
        "timothy2",
        "titus",
        "philemon",
        "hebrews",
        "james",
        "peter1",
        "peter2",
        "john1",
        "john2",
        "john3",
        "jude",
        "revelation",
    ]
    template = templates[int(book)]
    result = await DIALOGS.getInputDialog(title="Search Item", text="Enter an item to search for:")
    if not result:
        return ""
    return f"//{template}/{select}/{result}" if result else ""

async def uba_bible(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_bible,
        options=options,
        descriptions=descriptions,
        title="Bible",
        text="Select a bible version to continue:"
    )
    if not select:
        return ""
    result = await get_reference()
    return f"//bible/{select}/{result}" if result else ""

async def uba_chapter(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_bible,
        options=options,
        descriptions=descriptions,
        title="Bible",
        text="Select a bible version to continue:"
    )
    if not select:
        return ""
    result = await get_reference(verse_reference=False)
    return f"//chapter/{select}/{result}" if result else ""

async def uba_compare(options, descriptions):
    select = await get_multiple_bibles(options, descriptions)
    if not select:
        return ""
    else:
        select = "_".join(select)
    result = await get_reference()
    return f"//uba/COMPARE:::{select}:::{result}" if result else ""

async def uba_compare_chapter(options, descriptions):
    select = await get_multiple_bibles(options, descriptions)
    if not select:
        return ""
    else:
        select = "_".join(select)
    result = await get_reference(verse_reference=False)
    return f"//uba/COMPARECHAPTER:::{select}:::{result}" if result else ""

async def uba_commentary(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_commentary,
        options=options,
        descriptions=descriptions,
        title="Bible Commentary",
        text="Select a commentary to continue:"
    )
    if not select:
        return ""
    result = await get_reference()
    return f"//commentary/{select}/{result}" if result else ""

async def uba_dictionary():
    result = await DIALOGS.getInputDialog(title="Search Dictionary", text="Enter a search item:")
    return f"//dictionary/{result.strip()}" if result and result.strip() else ""

async def uba_parallel():
    result = await DIALOGS.getInputDialog(title="Search Bible Parallel Passages", text="Enter a search item:")
    return f"//parallel/{result.strip()}" if result and result.strip() else ""

async def uba_promise():
    result = await DIALOGS.getInputDialog(title="Search Bible Promises", text="Enter a search item:")
    return f"//promise/{result.strip()}" if result and result.strip() else ""

async def uba_topic():
    result = await DIALOGS.getInputDialog(title="Search Bible Topics", text="Enter a search item:")
    return f"//topic/{result.strip()}" if result and result.strip() else ""

async def uba_name():
    result = await DIALOGS.getInputDialog(title="Search Bible Names", text="Enter a search item:")
    return f"//name/{result.strip()}" if result and result.strip() else ""

async def uba_character():
    result = await DIALOGS.getInputDialog(title="Search Bible Characters", text="Enter a search item:")
    return f"//character/{result.strip()}" if result and result.strip() else ""

async def uba_location():
    result = await DIALOGS.getInputDialog(title="Search Bible Locations", text="Enter a search item:")
    return f"//location/{result.strip()}" if result and result.strip() else ""

async def uba_encyclopedia(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_encyclopedia,
        options=options,
        descriptions=descriptions,
        title="Encyclopedia",
        text="Select one of them to continue:"
    )
    if not select:
        return ""
    result = await DIALOGS.getInputDialog(title=f"Search Encyclopedia - {select}", text="Enter a search item:")
    return f"//encyclopedia/{select}/{result.strip()}" if result and result.strip() else ""

async def uba_lexicon(options):
    select = await DIALOGS.getValidOptions(
        default=config.default_lexicon,
        options=options,
        title="Lexicon",
        text="Select one of them to continue:"
    )
    if not select:
        return ""
    result = await DIALOGS.getInputDialog(title=f"Search Lexicon - {select}", text="Enter a search item:")
    return f"//lexicon/{select}/{result.strip()}" if result and result.strip() else ""

# Configure default modules

async def uba_default_bible(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_bible,
        options=options,
        descriptions=descriptions,
        title="Configure Default Bible",
        text="Select a bible version:"
    )
    return select if select else ""

async def uba_default_commentary(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_commentary,
        options=options,
        descriptions=descriptions,
        title="Configure Default Commentary",
        text="Select a commentary:"
    )
    return select if select else ""

async def uba_default_encyclopedia(options, descriptions):
    select = await DIALOGS.getValidOptions(
        default=config.default_encyclopedia,
        options=options,
        descriptions=descriptions,
        title="Configure Default Encyclopedia",
        text="Select an encyclopedia:"
    )
    return select if select else ""

async def uba_default_lexicon(options):
    select = await DIALOGS.getValidOptions(
        default=config.default_lexicon,
        options=options,
        descriptions=descriptions,
        title="Configure Default Lexicon",
        text="Select a lexicon:"
    )
    return select if select else ""


import logging, json, os
from fastmcp.server.auth.providers.jwt import StaticTokenVerifier
from fastmcp.server.auth.providers.jwt import JWTVerifier
from fastmcp import FastMCP
from fastmcp.prompts.prompt import PromptMessage, TextContent
from agentmake import agentmake, DEVELOPER_MODE, readTextFile
from biblemate import BIBLEMATEDATA, AGENTMAKE_CONFIG, config
from biblemate.uba.bible import search_bible
from biblemate.uba.api import run_uba_api
from typing import List, Dict, Any, Union

# configure backend
AGENTMAKE_CONFIG["backend"] = config.backend

# Configure logging before creating the FastMCP server
logging.basicConfig(format="[%(levelname)s]: %(message)s", level=logging.ERROR)

BIBLEMATE_STATIC_TOKEN = os.getenv("BIBLEMATE_STATIC_TOKEN")
BIBLEMATE_MCP_PUBLIC_KEY = os.getenv("BIBLEMATE_MCP_PUBLIC_KEY")

verifier = StaticTokenVerifier(
    tokens={
        BIBLEMATE_STATIC_TOKEN: {
            "client_id": "biblemate-ai",
            "scopes": ["read:data", "write:data", "admin:users"]
        },
    },
    required_scopes=["read:data"]
) if BIBLEMATE_STATIC_TOKEN else JWTVerifier(
    public_key=BIBLEMATE_MCP_PUBLIC_KEY,
    issuer=os.getenv("BIBLEMATE_MCP_ISSUER"),
    audience=os.getenv("BIBLEMATE_MCP_AUDIENCE")
) if BIBLEMATE_MCP_PUBLIC_KEY else None

mcp = FastMCP(name="BibleMate AI", auth=verifier)

def getResponse(messages:list) -> str:
    return messages[-1].get("content") if messages and "content" in messages[-1] else "Error!"

@mcp.resource("resource://info")
def info() -> str:
    """Display BibleMate AI information"""
    info = "BibleMate AI " + readTextFile(os.path.join(os.path.dirname(os.path.realpath(__file__)), "version.txt"))
    info += "\n\nSource: https://github.com/eliranwong/biblemate\nDeveloper: Eliran Wong"
    return info

@mcp.resource("uba://{command}")
def uba(command:str) -> str:
    """Execute an UBA command; a valid UBA command must be given, e.g. `//uba/John 3:16`; do not use this prompt if you are not sure what you are doing"""
    return run_uba_api(command)

if DEVELOPER_MODE:
    @mcp.resource("resource://audio")
    def audio() -> str:
        """Bible Audio"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["bibleAudioModules"]])

@mcp.resource("resource://bibles")
def bibles() -> dict:
    """Bibles; prompt examples: `//bible/John 3:16-18`, `//bible/KJV/John 3:16-18; Deut 6:4`"""
    resources = json.loads(run_uba_api(".resources"))
    return dict(zip(resources["bibleListAbb"], resources["bibleList"]))

@mcp.resource("bible://{module}/{reference}")
def bible(module:str, reference:str) -> str:
    """Bible; prompt examples: `//bible/John 3:16-18`, `//bible/KJV/John 3:16-18; Deut 6:4`"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    reference = BibleVerseParser(False).extractAllReferencesReadable(reference)
    if not reference:
        return "Please provide a valid Bible reference to complete your request."
    return f"{reference}\n\n"+run_uba_api(f"BIBLE:::{module}:::{reference}")

@mcp.resource("chapter://{module}/{reference}")
def chapter(module:str, reference:str) -> str:
    """retrieve a whole Bible chapter; bible chapter reference must be given, e.g. John 3"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    import re
    refs = re.sub("[Cc]hapter ([0-9])", r"\1", reference)
    refs = BibleVerseParser(False).extractAllReferencesReadable(refs)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    output = []
    for ref in refs.split("; "):
        output.append("# "+re.sub("\n([0-9])", r"\n* \1", run_uba_api(f"CHAPTER:::{module}:::{ref}")).replace("\n# ", "\n## "))
    return "\n\n".join(output)

@mcp.resource("resource://commentaries")
def commentaries() -> dict:
    """Commentaries; prompt examples: `//commentary/John 3:16`, `//commentary/CBSC/John 3:16`"""
    resources = json.loads(run_uba_api(".resources"))
    return dict(zip(resources["commentaryListAbb"], resources["commentaryList"]))

@mcp.resource("commentary://{module}/{reference}")
def commentary(module:str, reference:str) -> str:
    """Commentary; prompt examples: `//commentary/John 3:16`, `//commentary/CBSC/John 3:16`"""
    return run_uba_api(f"COMMENTARY:::{module}:::{reference}")

@mcp.resource("treasury://{reference}")
def treasury(reference:str) -> str:
    """Treasury of Scripture Knowledge (Enhance); prompt examples: `//treasury/John 3:16`, `//treasury/Deut 6:4`"""
    return run_uba_api(f"TSKE:::{config.default_bible}:::{reference}")

if DEVELOPER_MODE:
    @mcp.resource("resource://data")
    def data() -> str:
        """Data; UBA command example: `DATA:::Bible Chronology`"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["dataList"]])

@mcp.resource("resource://dictionaries")
def dictionaries() -> dict:
    """Dictionaries; prompt examples: `//dictionary/Jesus`, `//dictionary/Israel`"""
    resources = json.loads(run_uba_api(".resources"))
    return dict(zip(resources["dictionaryListAbb"], resources["dictionaryList"]))

dictionary_db = os.path.join(BIBLEMATEDATA, "dictionary.db")
if os.path.isfile(dictionary_db):
    @mcp.resource("dictionary://{query}")
    def dictionary(query:str) -> Union[str, list]:
        """Dictionary; prompt examples: `//dictionary/Jesus`, `//dictionary/Israel`"""
        from biblemate.uba.search import UBASearches
        dictionary_db = os.path.join(BIBLEMATEDATA, "dictionary.db")
        return UBASearches.search_data(
            db_file=dictionary_db,
            sql_table="Dictionary",
            query=query,
            top_k=config.max_semantic_matches,
        )

if DEVELOPER_MODE:
    @mcp.resource("resource://docs")
    def docs() -> str:
        """Documents"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["docxList"]])

@mcp.resource("resource://encyclopedias")
def encyclopedias() -> dict:
    """Encyclopedias; prompt examples: `//encyclopedia/Jesus`, `//encyclopedia/ISB/Jesus`"""
    resources = json.loads(run_uba_api(".resources"))
    return dict(zip(resources["encyclopediaListAbb"], resources["encyclopediaList"]))

encyclopedia_db = os.path.join(BIBLEMATEDATA, "encyclopedia.db")
if os.path.isfile(encyclopedia_db):
    @mcp.resource("encyclopedia://{module}/{query}")
    def encyclopedia(module: str, query:str) -> Union[str, list]:
        """Encyclopedia; prompt examples: `//encyclopedia/Jesus`, `//encyclopedia/ISB/Jesus`"""
        from biblemate.uba.search import UBASearches
        encyclopedia_db = os.path.join(BIBLEMATEDATA, "encyclopedia.db")
        return UBASearches.search_data(
            db_file=encyclopedia_db,
            sql_table=module,
            query=query,
            top_k=config.max_semantic_matches,
        )

if DEVELOPER_MODE:
    @mcp.resource("resource://epubs")
    def epubs() -> str:
        """EPUBs"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["epubList"]])

@mcp.resource("resource://lexicons")
def lexicons() -> str:
    """Lexicons; prompt examples: `//lexicon/G25`, `//lexicon/TBESH/G25`, `//lexicon/TBESH/H3478`"""
    resources = json.loads(run_uba_api(".resources"))
    return "\n".join([f"- `{r}`" for r in resources["lexiconList"]])

@mcp.resource("lexicon://{module}/{entry}")
def lexicon(module:str, entry:str) -> str:
    """Lexicon; ; prompt examples: `//lexicon/G25`, `//lexicon/TBESH/G25`, `//lexicon/TBESH/H3478`"""
    import re
    command = f"LEXICON:::{module}:::{entry}"
    content = run_uba_api(command)
    content = content.replace("\n", "\n- ")
    content = re.sub("\n.*?More lexicons.*?\n", "\n", content)
    return content.replace(" [ search ]", "")

if DEVELOPER_MODE:
    @mcp.resource("resource://references")
    def references() -> str:
        """Reference Books"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["referenceBookList"]])

    @mcp.resource("resource://pdfs")
    def pdfs() -> str:
        """PDFs"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["pdfList"]])

    @mcp.resource("resource://searchtools")
    def searchtools() -> str:
        """Search Tools"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["searchToolList"]])

@mcp.resource("resource://strongs")
def strongs() -> str:
    """Strong's Bibles; UBA command example: `BIBLE:::KJVx:::John 3:16`"""
    resources = json.loads(run_uba_api(".resources"))
    return "\n".join([f"- `{r}`" for r in resources["strongBibleListAbb"]])

if DEVELOPER_MODE:
    @mcp.resource("resource://thirddicts")
    def thirddicts() -> str:
        """Third-Party Dictionaries; UBA command examples: `SEARCHTHIRDDICTIONARY:::faith`, `SEARCHTHIRDDICTIONARY:::webster:::faith`"""
        resources = json.loads(run_uba_api(".resources"))
        return "\n".join([f"- `{r}`" for r in resources["thirdPartyDictionaryList"]])

@mcp.resource("resource://topics")
def topics() -> dict:
    """Topical Collections; prompt examples: `//topic/faith`, `//topic/hope`, `//topic/love`"""
    resources = json.loads(run_uba_api(".resources"))
    return dict(zip(resources["topicListAbb"], resources["topicList"]))

collection_db = os.path.join(BIBLEMATEDATA, "collection.db")
if os.path.isfile(collection_db):
    @mcp.resource("parallel://{query}")
    def parallel(query:str) -> Union[str, list]:
        """Bible Parallels; prompt examples: `//parallel/baptism`, `//parallel/light`, `//parallel/sermon`"""
        from biblemate.uba.search import UBASearches
        collection_db = os.path.join(BIBLEMATEDATA, "collection.db")
        return UBASearches.search_data(
            db_file=collection_db,
            sql_table="PARALLEL",
            query=query,
            top_k=config.max_semantic_matches,
        )
    @mcp.resource("promise://{query}")
    def promise(query:str) -> Union[str, list]:
        """Bible Promises; prompt examples: `//promise/faith`, `//promise/hope`, `//promise/love`"""
        from biblemate.uba.search import UBASearches
        collection_db = os.path.join(BIBLEMATEDATA, "collection.db")
        return UBASearches.search_data(
            db_file=collection_db,
            sql_table="PROMISES",
            query=query,
            top_k=config.max_semantic_matches,
        )

topic_db = os.path.join(BIBLEMATEDATA, "exlb.db")
if os.path.isfile(topic_db):
    @mcp.resource("topic://{query}")
    def topic(query:str) -> Union[str, list]:
        """Topical Studies; prompt examples: `//topic/faith`, `//topic/hope`, `//topic/love`"""
        from biblemate.uba.search import UBASearches
        topic_db = os.path.join(BIBLEMATEDATA, "exlb.db")
        return UBASearches.search_data(
            db_file=topic_db,
            sql_table="exlbt",
            query=query,
            top_k=config.max_semantic_matches,
        )
    @mcp.resource("name://{query}")
    def name(query:str) -> Union[str, list]:
        """Bible Names; prompt examples: `//name/Jesus`, `//name/Bethlehem`"""
        from biblemate.uba.search import UBASearches
        topic_db = os.path.join(BIBLEMATEDATA, "exlb.db")
        return UBASearches.search_data(
            db_file=topic_db,
            sql_table="exlbn",
            query=query,
            top_k=config.max_semantic_matches,
        )
    @mcp.resource("character://{query}")
    def character(query:str) -> Union[str, list]:
        """Character Studies; prompt examples: `//character/Jesus`, `//character/Samuel`, `//topic/John`"""
        from biblemate.uba.search import UBASearches
        topic_db = os.path.join(BIBLEMATEDATA, "exlb.db")
        return UBASearches.search_data(
            db_file=topic_db,
            sql_table="exlbp",
            query=query,
            top_k=config.max_semantic_matches,
        )
    @mcp.resource("location://{query}")
    def location(query:str) -> Union[str, list]:
        """Location Studies; prompt examples: `//location/Jerusalem`, `//location/Bethel`, `//location/Bethlehem`"""
        from biblemate.uba.search import UBASearches
        topic_db = os.path.join(BIBLEMATEDATA, "exlb.db")
        return UBASearches.search_data(
            db_file=topic_db,
            sql_table="exlbl",
            query=query,
            top_k=config.max_semantic_matches,
        )

@mcp.resource("search://{module}/{request}")
def search(module:str, request:str) -> str:
    """search the whole bible; search string must be given"""
    global search_bible
    return search_bible(request=request, module=module, search_request=True)

@mcp.resource("genesis://{module}/{request}")
def genesis(module:str, request:str) -> str:
    """search the book of Genesis only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=1, module=module, search_request=True)

@mcp.resource("exodus://{module}/{request}")
def exodus(module:str, request:str) -> str:
    """search the book of Exodus only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=2, module=module, search_request=True)

@mcp.resource("leviticus://{module}/{request}")
def leviticus(module:str, request:str) -> str:
    """search the book of Leviticus only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=3, module=module, search_request=True)

@mcp.resource("numbers://{module}/{request}")
def numbers(module:str, request:str) -> str:
    """search the book of Numbers only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=4, module=module, search_request=True)

@mcp.resource("deuteronomy://{module}/{request}")
def deuteronomy(module:str, request:str) -> str:
    """search the book of Deuteronomy only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=5, module=module, search_request=True)

@mcp.resource("joshua://{module}/{request}")
def joshua(module:str, request:str) -> str:
    """search the book of Joshua only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=6, module=module, search_request=True)

@mcp.resource("judges://{module}/{request}")
def judges(module:str, request:str) -> str:
    """search the book of Judges only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=7, module=module, search_request=True)

@mcp.resource("ruth://{module}/{request}")
def ruth(module:str, request:str) -> str:
    """search the book of Ruth only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=8, module=module, search_request=True)

@mcp.resource("samuel1://{module}/{request}")
def samuel1(module:str, request:str) -> str:
    """search the book of 1 Samuel only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=9, module=module, search_request=True)

@mcp.resource("samuel2://{module}/{request}")
def samuel2(module:str, request:str) -> str:
    """search the book of 2 Samuel only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=10, module=module, search_request=True)

@mcp.resource("kings1://{module}/{request}")
def kings1(module:str, request:str) -> str:
    """search the book of 1 Kings only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=11, module=module, search_request=True)

@mcp.resource("kings2://{module}/{request}")
def kings2(module:str, request:str) -> str:
    """search the book of 2 Kings only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=12, module=module, search_request=True)

@mcp.resource("chronicles1://{module}/{request}")
def chronicles1(module:str, request:str) -> str:
    """search the book of 1 Chronicles only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=13, module=module, search_request=True)

@mcp.resource("chronicles2://{module}/{request}")
def chronicles2(module:str, request:str) -> str:
    """search the book of 2 Chronicles only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=14, module=module, search_request=True)

@mcp.resource("ezra://{module}/{request}")
def ezra(module:str, request:str) -> str:
    """search the book of Ezra only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=15, module=module, search_request=True)

@mcp.resource("nehemiah://{module}/{request}")
def nehemiah(module:str, request:str) -> str:
    """search the book of Nehemiah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=16, module=module, search_request=True)

@mcp.resource("esther://{module}/{request}")
def esther(module:str, request:str) -> str:
    """search the book of Esther only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=17, module=module, search_request=True)

@mcp.resource("job://{module}/{request}")
def job(module:str, request:str) -> str:
    """search the book of Job only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=18, module=module, search_request=True)

@mcp.resource("psalms://{module}/{request}")
def psalms(module:str, request:str) -> str:
    """search the book of Psalms only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=19, module=module, search_request=True)

@mcp.resource("proverbs://{module}/{request}")
def proverbs(module:str, request:str) -> str:
    """search the book of Proverbs only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=20, module=module, search_request=True)

@mcp.resource("ecclesiastes://{module}/{request}")
def ecclesiastes(module:str, request:str) -> str:
    """search the book of Ecclesiastes only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=21, module=module, search_request=True)

@mcp.resource("songs://{module}/{request}")
def songs(module:str, request:str) -> str:
    """search the book of Song of Songs only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=22, module=module, search_request=True)

@mcp.resource("isaiah://{module}/{request}")
def isaiah(module:str, request:str) -> str:
    """search the book of Isaiah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=23, module=module, search_request=True)

@mcp.resource("jeremiah://{module}/{request}")
def jeremiah(module:str, request:str) -> str:
    """search the book of Jeremiah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=24, module=module, search_request=True)

@mcp.resource("lamentations://{module}/{request}")
def lamentations(module:str, request:str) -> str:
    """search the book of Lamentations only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=25, module=module, search_request=True)

@mcp.resource("ezekiel://{module}/{request}")
def ezekiel(module:str, request:str) -> str:
    """search the book of Ezekiel only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=26, module=module, search_request=True)

@mcp.resource("daniel://{module}/{request}")
def daniel(module:str, request:str) -> str:
    """search the book of Daniel only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=27, module=module, search_request=True)

@mcp.resource("hosea://{module}/{request}")
def hosea(module:str, request:str) -> str:
    """search the book of Hosea only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=28, module=module, search_request=True)

@mcp.resource("joel://{module}/{request}")
def joel(module:str, request:str) -> str:
    """search the book of Joel only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=29, module=module, search_request=True)

@mcp.resource("amos://{module}/{request}")
def amos(module:str, request:str) -> str:
    """search the book of Amos only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=30, module=module, search_request=True)

@mcp.resource("obadiah://{module}/{request}")
def obadiah(module:str, request:str) -> str:
    """search the book of Obadiah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=31, module=module, search_request=True)

@mcp.resource("jonah://{module}/{request}")
def jonah(module:str, request:str) -> str:
    """search the book of Jonah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=32, module=module, search_request=True)

@mcp.resource("micah://{module}/{request}")
def micah(module:str, request:str) -> str:
    """search the book of Micah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=33, module=module, search_request=True)

@mcp.resource("nahum://{module}/{request}")
def nahum(module:str, request:str) -> str:
    """search the book of Nahum only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=34, module=module, search_request=True)

@mcp.resource("habakkuk://{module}/{request}")
def habakkuk(module:str, request:str) -> str:
    """search the book of Habakkuk only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=35, module=module, search_request=True)

@mcp.resource("zephaniah://{module}/{request}")
def zephaniah(module:str, request:str) -> str:
    """search the book of Zephaniah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=36, module=module, search_request=True)

@mcp.resource("haggai://{module}/{request}")
def haggai(module:str, request:str) -> str:
    """search the book of Haggai only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=37, module=module, search_request=True)

@mcp.resource("zechariah://{module}/{request}")
def zechariah(module:str, request:str) -> str:
    """search the book of Zechariah only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=38, module=module, search_request=True)

@mcp.resource("malachi://{module}/{request}")
def malachi(module:str, request:str) -> str:
    """search the book of Malachi only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=39, module=module, search_request=True)

@mcp.resource("matthew://{module}/{request}")
def matthew(module:str, request:str) -> str:
    """search the book of Matthew only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=40, module=module, search_request=True)

@mcp.resource("mark://{module}/{request}")
def mark(module:str, request:str) -> str:
    """search the book of Mark only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=41, module=module, search_request=True)

@mcp.resource("luke://{module}/{request}")
def luke(module:str, request:str) -> str:
    """search the book of Luke only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=42, module=module, search_request=True)

@mcp.resource("john://{module}/{request}")
def john(module:str, request:str) -> str:
    """search the book of John only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=43, module=module, search_request=True)

@mcp.resource("acts://{module}/{request}")
def acts(module:str, request:str) -> str:
    """search the book of Acts only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=44, module=module, search_request=True)

@mcp.resource("romans://{module}/{request}")
def romans(module:str, request:str) -> str:
    """search the book of Romans only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=45, module=module, search_request=True)

@mcp.resource("corinthians1://{module}/{request}")
def corinthians1(module:str, request:str) -> str:
    """search the book of 1 Corinthians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=46, module=module, search_request=True)

@mcp.resource("corinthians2://{module}/{request}")
def corinthians2(module:str, request:str) -> str:
    """search the book of 2 Corinthians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=47, module=module, search_request=True)

@mcp.resource("galatians://{module}/{request}")
def galatians(module:str, request:str) -> str:
    """search the book of Galatians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=48, module=module, search_request=True)

@mcp.resource("ephesians://{module}/{request}")
def ephesians(module:str, request:str) -> str:
    """search the book of Ephesians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=49, module=module, search_request=True)

@mcp.resource("philippians://{module}/{request}")
def philippians(module:str, request:str) -> str:
    """search the book of Philippians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=50, module=module, search_request=True)

@mcp.resource("colossians://{module}/{request}")
def colossians(module:str, request:str) -> str:
    """search the book of Colossians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=51, module=module, search_request=True)

@mcp.resource("thessalonians1://{module}/{request}")
def thessalonians1(module:str, request:str) -> str:
    """search the book of 1 Thessalonians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=52, module=module, search_request=True)

@mcp.resource("thessalonians2://{module}/{request}")
def thessalonians2(module:str, request:str) -> str:
    """search the book of 2 Thessalonians only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=53, module=module, search_request=True)

@mcp.resource("timothy1://{module}/{request}")
def timothy1(module:str, request:str) -> str:
    """search the book of 1 Timothy only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=54, module=module, search_request=True)

@mcp.resource("timothy2://{module}/{request}")
def timothy2(module:str, request:str) -> str:
    """search the book of 2 Timothy only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=55, module=module, search_request=True)

@mcp.resource("titus://{module}/{request}")
def titus(module:str, request:str) -> str:
    """search the book of Titus only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=56, module=module, search_request=True)

@mcp.resource("philemon://{module}/{request}")
def philemon(module:str, request:str) -> str:
    """search the book of Philemon only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=57, module=module, search_request=True)

@mcp.resource("hebrews://{module}/{request}")
def hebrews(module:str, request:str) -> str:
    """search the book of Hebrews only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=58, module=module, search_request=True)

@mcp.resource("james://{module}/{request}")
def james(module:str, request:str) -> str:
    """search the book of James only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=59, module=module, search_request=True)

@mcp.resource("peter1://{module}/{request}")
def peter1(module:str, request:str) -> str:
    """search the book of 1 Peter only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=60, module=module, search_request=True)

@mcp.resource("peter2://{module}/{request}")
def peter2(module:str, request:str) -> str:
    """search the book of 2 Peter only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=61, module=module, search_request=True)

@mcp.resource("john1://{module}/{request}")
def john1(module:str, request:str) -> str:
    """search the book of 1 John only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=62, module=module, search_request=True)

@mcp.resource("john2://{module}/{request}")
def john2(module:str, request:str) -> str:
    """search the book of 2 John only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=63, module=module, search_request=True)

@mcp.resource("john3://{module}/{request}")
def john3(module:str, request:str) -> str:
    """search the book of 3 John only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=64, module=module, search_request=True)

@mcp.resource("jude://{module}/{request}")
def jude(module:str, request:str) -> str:
    """search the book of Jude only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=65, module=module, search_request=True)

@mcp.resource("revelation://{module}/{request}")
def revelation(module:str, request:str) -> str:
    """search the book of Revelation only; search string must be given"""
    global search_bible
    return search_bible(request=request, book=66, module=module, search_request=True)

@mcp.tool
def search_the_whole_bible(request:str) -> str:
    """search the whole bible; search string must be given"""
    global search_bible
    return search_bible(request)

@mcp.tool
def search_genesis_only(request:str) -> str:
    """search the book of Genesis only; search string must be given"""
    global search_bible
    return search_bible(request, book=1)

@mcp.tool
def search_exodus_only(request:str) -> str:
    """search the book of Exodus only; search string must be given"""
    global search_bible
    return search_bible(request, book=2)

@mcp.tool
def search_leviticus_only(request:str) -> str:
    """search the book of Leviticus only; search string must be given"""
    global search_bible
    return search_bible(request, book=3)

@mcp.tool
def search_numbers_only(request:str) -> str:
    """search the book of Numbers only; search string must be given"""
    global search_bible
    return search_bible(request, book=4)

@mcp.tool
def search_deuteronomy_only(request:str) -> str:
    """search the book of Deuteronomy only; search string must be given"""
    global search_bible
    return search_bible(request, book=5)

@mcp.tool
def search_joshua_only(request:str) -> str:
    """search the book of Joshua only; search string must be given"""
    global search_bible
    return search_bible(request, book=6)

@mcp.tool
def search_judges_only(request:str) -> str:
    """search the book of Judges only; search string must be given"""
    global search_bible
    return search_bible(request, book=7)

@mcp.tool
def search_ruth_only(request:str) -> str:
    """search the book of Ruth only; search string must be given"""
    global search_bible
    return search_bible(request, book=8)

@mcp.tool
def search_1_samuel_only(request:str) -> str:
    """search the book of 1 Samuel only; search string must be given"""
    global search_bible
    return search_bible(request, book=9)

@mcp.tool
def search_2_samuel_only(request:str) -> str:
    """search the book of 2 Samuel only; search string must be given"""
    global search_bible
    return search_bible(request, book=10)

@mcp.tool
def search_1_kings_only(request:str) -> str:
    """search the book of 1 Kings only; search string must be given"""
    global search_bible
    return search_bible(request, book=11)

@mcp.tool
def search_2_kings_only(request:str) -> str:
    """search the book of 2 Kings only; search string must be given"""
    global search_bible
    return search_bible(request, book=12)

@mcp.tool
def search_1_chronicles_only(request:str) -> str:
    """search the book of 1 Chronicles only; search string must be given"""
    global search_bible
    return search_bible(request, book=13)

@mcp.tool
def search_2_chronicles_only(request:str) -> str:
    """search the book of 2 Chronicles only; search string must be given"""
    global search_bible
    return search_bible(request, book=14)

@mcp.tool
def search_ezra_only(request:str) -> str:
    """search the book of Ezra only; search string must be given"""
    global search_bible
    return search_bible(request, book=15)

@mcp.tool
def search_nehemiah_only(request:str) -> str:
    """search the book of Nehemiah only; search string must be given"""
    global search_bible
    return search_bible(request, book=16)

@mcp.tool
def search_esther_only(request:str) -> str:
    """search the book of Esther only; search string must be given"""
    global search_bible
    return search_bible(request, book=17)

@mcp.tool
def search_job_only(request:str) -> str:
    """search the book of Job only; search string must be given"""
    global search_bible
    return search_bible(request, book=18)

@mcp.tool
def search_psalms_only(request:str) -> str:
    """search the book of Psalms only; search string must be given"""
    global search_bible
    return search_bible(request, book=19)

@mcp.tool
def search_proverbs_only(request:str) -> str:
    """search the book of Proverbs only; search string must be given"""
    global search_bible
    return search_bible(request, book=20)

@mcp.tool
def search_ecclesiastes_only(request:str) -> str:
    """search the book of Ecclesiastes only; search string must be given"""
    global search_bible
    return search_bible(request, book=21)

@mcp.tool
def search_song_of_songs_only(request:str) -> str:
    """search the book of Song of Songs only; search string must be given"""
    global search_bible
    return search_bible(request, book=22)

@mcp.tool
def search_isaiah_only(request:str) -> str:
    """search the book of Isaiah only; search string must be given"""
    global search_bible
    return search_bible(request, book=23)

@mcp.tool
def search_jeremiah_only(request:str) -> str:
    """search the book of Jeremiah only; search string must be given"""
    global search_bible
    return search_bible(request, book=24)

@mcp.tool
def search_lamentations_only(request:str) -> str:
    """search the book of Lamentations only; search string must be given"""
    global search_bible
    return search_bible(request, book=25)

@mcp.tool
def search_ezekiel_only(request:str) -> str:
    """search the book of Ezekiel only; search string must be given"""
    global search_bible
    return search_bible(request, book=26)

@mcp.tool
def search_daniel_only(request:str) -> str:
    """search the book of Daniel only; search string must be given"""
    global search_bible
    return search_bible(request, book=27)

@mcp.tool
def search_hosea_only(request:str) -> str:
    """search the book of Hosea only; search string must be given"""
    global search_bible
    return search_bible(request, book=28)

@mcp.tool
def search_joel_only(request:str) -> str:
    """search the book of Joel only; search string must be given"""
    global search_bible
    return search_bible(request, book=29)

@mcp.tool
def search_amos_only(request:str) -> str:
    """search the book of Amos only; search string must be given"""
    global search_bible
    return search_bible(request, book=30)

@mcp.tool
def search_obadiah_only(request:str) -> str:
    """search the book of Obadiah only; search string must be given"""
    global search_bible
    return search_bible(request, book=31)

@mcp.tool
def search_jonah_only(request:str) -> str:
    """search the book of Jonah only; search string must be given"""
    global search_bible
    return search_bible(request, book=32)

@mcp.tool
def search_micah_only(request:str) -> str:
    """search the book of Micah only; search string must be given"""
    global search_bible
    return search_bible(request, book=33)

@mcp.tool
def search_nahum_only(request:str) -> str:
    """search the book of Nahum only; search string must be given"""
    global search_bible
    return search_bible(request, book=34)

@mcp.tool
def search_habakkuk_only(request:str) -> str:
    """search the book of Habakkuk only; search string must be given"""
    global search_bible
    return search_bible(request, book=35)

@mcp.tool
def search_zephaniah_only(request:str) -> str:
    """search the book of Zephaniah only; search string must be given"""
    global search_bible
    return search_bible(request, book=36)

@mcp.tool
def search_haggai_only(request:str) -> str:
    """search the book of Haggai only; search string must be given"""
    global search_bible
    return search_bible(request, book=37)

@mcp.tool
def search_zechariah_only(request:str) -> str:
    """search the book of Zechariah only; search string must be given"""
    global search_bible
    return search_bible(request, book=38)

@mcp.tool
def search_malachi_only(request:str) -> str:
    """search the book of Malachi only; search string must be given"""
    global search_bible
    return search_bible(request, book=39)

@mcp.tool
def search_matthew_only(request:str) -> str:
    """search the book of Matthew only; search string must be given"""
    global search_bible
    return search_bible(request, book=40)

@mcp.tool
def search_mark_only(request:str) -> str:
    """search the book of Mark only; search string must be given"""
    global search_bible
    return search_bible(request, book=41)

@mcp.tool
def search_luke_only(request:str) -> str:
    """search the book of Luke only; search string must be given"""
    global search_bible
    return search_bible(request, book=42)

@mcp.tool
def search_john_only(request:str) -> str:
    """search the book of John only; search string must be given"""
    global search_bible
    return search_bible(request, book=43)

@mcp.tool
def search_acts_only(request:str) -> str:
    """search the book of Acts only; search string must be given"""
    global search_bible
    return search_bible(request, book=44)

@mcp.tool
def search_romans_only(request:str) -> str:
    """search the book of Romans only; search string must be given"""
    global search_bible
    return search_bible(request, book=45)

@mcp.tool
def search_1_corinthians_only(request:str) -> str:
    """search the book of 1 Corinthians only; search string must be given"""
    global search_bible
    return search_bible(request, book=46)

@mcp.tool
def search_2_corinthians_only(request:str) -> str:
    """search the book of 2 Corinthians only; search string must be given"""
    global search_bible
    return search_bible(request, book=47)

@mcp.tool
def search_galatians_only(request:str) -> str:
    """search the book of Galatians only; search string must be given"""
    global search_bible
    return search_bible(request, book=48)

@mcp.tool
def search_ephesians_only(request:str) -> str:
    """search the book of Ephesians only; search string must be given"""
    global search_bible
    return search_bible(request, book=49)

@mcp.tool
def search_philippians_only(request:str) -> str:
    """search the book of Philippians only; search string must be given"""
    global search_bible
    return search_bible(request, book=50)

@mcp.tool
def search_colossians_only(request:str) -> str:
    """search the book of Colossians only; search string must be given"""
    global search_bible
    return search_bible(request, book=51)

@mcp.tool
def search_1_thessalonians_only(request:str) -> str:
    """search the book of 1 Thessalonians only; search string must be given"""
    global search_bible
    return search_bible(request, book=52)

@mcp.tool
def search_2_thessalonians_only(request:str) -> str:
    """search the book of 2 Thessalonians only; search string must be given"""
    global search_bible
    return search_bible(request, book=53)

@mcp.tool
def search_1_timothy_only(request:str) -> str:
    """search the book of 1 Timothy only; search string must be given"""
    global search_bible
    return search_bible(request, book=54)

@mcp.tool
def search_2_timothy_only(request:str) -> str:
    """search the book of 2 Timothy only; search string must be given"""
    global search_bible
    return search_bible(request, book=55)

@mcp.tool
def search_titus_only(request:str) -> str:
    """search the book of Titus only; search string must be given"""
    global search_bible
    return search_bible(request, book=56)

@mcp.tool
def search_philemon_only(request:str) -> str:
    """search the book of Philemon only; search string must be given"""
    global search_bible
    return search_bible(request, book=57)

@mcp.tool
def search_hebrews_only(request:str) -> str:
    """search the book of Hebrews only; search string must be given"""
    global search_bible
    return search_bible(request, book=58)

@mcp.tool
def search_james_only(request:str) -> str:
    """search the book of James only; search string must be given"""
    global search_bible
    return search_bible(request, book=59)

@mcp.tool
def search_1_peter_only(request:str) -> str:
    """search the book of 1 Peter only; search string must be given"""
    global search_bible
    return search_bible(request, book=60)

@mcp.tool
def search_2_peter_only(request:str) -> str:
    """search the book of 2 Peter only; search string must be given"""
    global search_bible
    return search_bible(request, book=61)

@mcp.tool
def search_1_john_only(request:str) -> str:
    """search the book of 1 John only; search string must be given"""
    global search_bible
    return search_bible(request, book=62)

@mcp.tool
def search_2_john_only(request:str) -> str:
    """search the book of 2 John only; search string must be given"""
    global search_bible
    return search_bible(request, book=63)

@mcp.tool
def search_3_john_only(request:str) -> str:
    """search the book of 3 John only; search string must be given"""
    global search_bible
    return search_bible(request, book=64)

@mcp.tool
def search_jude_only(request:str) -> str:
    """search the book of Jude only; search string must be given"""
    global search_bible
    return search_bible(request, book=65)

@mcp.tool
def search_revelation_only(request:str) -> str:
    """search the book of Revelation only; search string must be given"""
    global search_bible
    return search_bible(request, book=66)

@mcp.tool
def compare_bible_translations(request:str) -> str:
    """compare Bible translations; bible verse reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'input_content_plugin': 'uba/every_single_ref', 'tool': 'uba/compare'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def retrieve_bible_study_indexes(request:str) -> str:
    """retrieve study reference entries on studying a particular bible verse; bible verse reference must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'input_content_plugin': 'uba/every_single_ref', 'tool': 'uba/index'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def retrieve_bible_cross_references(request:str) -> str:
    """retrieve cross-references of Bible verses; bible verse reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'input_content_plugin': 'uba/every_single_ref', 'tool': 'uba/xref'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def retrieve_pointed_hebrew_or_accented_greek_bible_verses(request:str) -> str:
    """retrieve Hebrew (with pointed vowels) or Greek (with accents) Bible verses; bible verse reference(s) must be given, e.g. John 3:16-17; single or multiple references accepted, e.g. Deut 6:4; Gen 1:26-27"""
    global agentmake, getResponse
    messages = agentmake(request, **{'tool': 'uba/ohgb'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def retrieve_hebrew_or_greek_bible_verses(request:str) -> str:
    """retrieve Hebrew (without pointed vowels) or Greek (without accents) Bible verses; bible verse reference(s) must be given, e.g. John 3:16-17; single or multiple references accepted, e.g. Deut 6:4; Gen 1:26-27"""
    global agentmake, getResponse
    messages = agentmake(request, **{'tool': 'uba/mob'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def retrieve_bible_verses(request:str) -> str:
    """retrieve Bible verses; bible verse reference(s) must be given, e.g. John 3:16-17; single or multiple references accepted, e.g. Deut 6:4; Gen 1:26-27"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    refs = BibleVerseParser(False).extractAllReferencesReadable(request)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    return run_uba_api(f"BIBLE:::{config.default_bible}:::{refs}")

@mcp.tool
def retrieve_verse_translations(request:str) -> str:
    """retrieve interlinear Hebrew or Greek, together with both literal and dynamic translations of inidividual bible verses; bible verse; bible verse reference(s) must be given, e.g. John 3:16-17; single or multiple references accepted, e.g. Deut 6:4; Gen 1:26-27"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    import re
    refs = BibleVerseParser(False).extractExhaustiveReferencesReadable(request)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    output = ""
    for ref in refs.split("; "):
        command = f"TRANSLATION:::{ref}"
        content = run_uba_api(command)
        output += content.replace("\n", "\n- ")
    return output

@mcp.tool
def retrieve_verse_discourse(request:str) -> str:
    """retrieve discourse analysis of inidividual bible verses; bible verse; bible verse reference(s) must be given, e.g. John 3:16-17; single or multiple references accepted, e.g. Deut 6:4; Gen 1:26-27"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    import re
    refs = BibleVerseParser(False).extractExhaustiveReferencesReadable(request)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    output = ""
    for ref in refs.split("; "):
        command = f"DISCOURSE:::{ref}"
        content = run_uba_api(command)
        output += content.replace("\n", "\n- ")
    return output

@mcp.tool
def retrieve_verse_morphology(request:str) -> str:
    """retrieve parsing and morphology of individual bible verses; bible verse reference(s) must be given, e.g. John 3:16-17; single or multiple references accepted, e.g. Deut 6:4; Gen 1:26-27"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    import re
    refs = BibleVerseParser(False).extractExhaustiveReferencesReadable(request)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    output = ""
    for ref in refs.split("; "):
        command = f"WORDS:::{ref}"
        morphology = run_uba_api(command, True)
        morphology = re.sub('''<[^<>]*?(READWORD|READLEXEME)(:::.*?)'">''', r'\n- [\1\2]\n- ', morphology)
        morphology = morphology.replace("audiotrack", "")
        morphology = morphology.replace("<div ", "\n### <div ")
        morphology = re.sub('''<[^<>]*? G(E[0-9]+?) (H[0-9]+?)"[^<>]*?>([^<>]*?)<[^<>]*?>''', r"\3 [\1] [\2]", morphology)
        morphology = re.sub('''<[^<>]*? (G[0-9]+?)"[^<>]*?>([^<>]*?)<[^<>]*?>''', r"\2 [\1]", morphology)
        morphology = morphology.replace("<heb>", "\n- <heb>")
        morphology = morphology.replace("<grk>", "\n- <grk>")
        morphology = morphology.replace("<br>", "\n- <br>")
        morphology = re.sub('<[^<>]*?>', '', morphology)
        output += "# "+morphology
    return output

@mcp.tool
def retrieve_bible_chapter(request:str) -> str:
    """retrieve a whole Bible chapter; bible chapter reference must be given, e.g. John 3"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    import re
    refs = re.sub("[Cc]hapter ([0-9])", r"\1", request)
    refs = BibleVerseParser(False).extractAllReferencesReadable(refs)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    output = []
    for ref in refs.split("; "):
        output.append("# "+re.sub("\n([0-9])", r"\n* \1", run_uba_api(f"CHAPTER:::{config.default_bible}:::{ref}")).replace("\n# ", "\n## "))
    return "\n\n".join(output)

@mcp.tool
def read_bible_commentary(request:str) -> str:
    """read bible commentary on individual bible verses; bible verse reference(s) must be given, like , like John 3:16 or John 3:16-18"""
    from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
    global agentmake, getResponse
    refs = BibleVerseParser(False).extractExhaustiveReferencesReadable(request)
    if not refs:
        return "Please provide a valid Bible reference to complete your request."
    output = []
    for ref in refs.split("; "):
        default_verse = run_uba_api(f"BIBLE:::{config.default_bible}:::{ref}")
        interlinear_verse = run_uba_api(f"BIBLE:::OHGBi:::{ref}")
        prompt = f"""# Write a detailed commentary on the following Bible verse:\n\n## {ref}\n{default_verse}\n\n##Interlinear (Hebrew/Greek with literal translation):\n{interlinear_verse}\n\nCommentary:"""
        messages = agentmake(prompt, system="biblemate/commentary", **AGENTMAKE_CONFIG)
        output.append(getResponse(messages))
    return f"# Commentary - {ref}\n\n"+"\n\n".join(output)

@mcp.tool
def refine_bible_translation(request:List[Dict[str, Any]]) -> str:
    """refine the translation of a Bible verse or passage"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/translate'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_pastor_prayer(request:List[Dict[str, Any]]) -> str:
    """write a prayer, out of a church pastor heart, based on user input"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/pray'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def ask_theologian(request:List[Dict[str, Any]]) -> str:
    """ask a theologian about the bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/theologian'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def quote_bible_verses(request:List[Dict[str, Any]]) -> str:
    """quote multiple bible verses in response to user request"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/quote'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def anyalyze_psalms(request:List[Dict[str, Any]]) -> str:
    """analyze the context and background of the Psalms in the bible; Psalm reference must be given, e.g. Psalm 23:1-3"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/david'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def ask_pastor(request:List[Dict[str, Any]]) -> str:
    """ask a church pastor about the bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/billy'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def ask_bible_scholar(request:List[Dict[str, Any]]) -> str:
    """ask a bible scholar about the bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'system': 'bible/scholar'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def explain_bible_meaning(request:List[Dict[str, Any]]) -> str:
    """Explain the meaning of the user-given content in reference to the Bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/meaning', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_new_testament_historical_context(request:List[Dict[str, Any]]) -> str:
    """write the Bible Historical Context of a New Testament passage in the bible; new testament bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/nt_context', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_questions(request:List[Dict[str, Any]]) -> str:
    """Write thought-provoking questions for bible study group discussion; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/questions', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_devotion(request:List[Dict[str, Any]]) -> str:
    """Write a devotion on a bible passage; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/devotion', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def translate_hebrew_bible_verse(request:List[Dict[str, Any]]) -> str:
    """Translate a Hebrew bible verse; Hebrew bible text must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/translate_hebrew', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_location_study(request:List[Dict[str, Any]]) -> str:
    """write comprehensive information on a bible location; a bible location name must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/location', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def translate_greek_bible_verse(request:List[Dict[str, Any]]) -> str:
    """Translate a Greek bible verse: Greek bible text must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/translate_greek', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def identify_bible_keywords(request:List[Dict[str, Any]]) -> str:
    """Identify bible key words from the user-given content"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/keywords', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def study_old_testament_themes(request:List[Dict[str, Any]]) -> str:
    """Study Bible Themes in a Old Testament passage; old testatment bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/ot_themes', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def study_new_testament_themes(request:List[Dict[str, Any]]) -> str:
    """Study Bible Themes in a New Testament passage; new testament bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/nt_themes', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_old_testament_highlights(request:List[Dict[str, Any]]) -> str:
    """Write Highlights in a Old Testament passage in the bible; old testament bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/ot_highligths', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_prayer(request:List[Dict[str, Any]]) -> str:
    """Write a prayer pertaining to the user content in reference to the Bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/prayer', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_short_bible_prayer(request:List[Dict[str, Any]]) -> str:
    """Write a short prayer, in one paragraph only, pertaining to the user content in reference to the Bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/short_prayer', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_character_study(request:List[Dict[str, Any]]) -> str:
    """Write comprehensive information on a given bible character in the bible; a bible character name must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/character', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_thought_progression(request:List[Dict[str, Any]]) -> str:
    """write Bible Thought Progression of a bible book / chapter / passage; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/flow', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def quote_bible_promises(request:List[Dict[str, Any]]) -> str:
    """Quote relevant Bible promises in response to user request"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/promises', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_chapter_summary(request:List[Dict[str, Any]]) -> str:
    """Write a detailed interpretation on a bible chapter; a bible chapter must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/chapter_summary', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_perspectives(request:List[Dict[str, Any]]) -> str:
    """Write biblical perspectives and principles in relation to the user content"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/perspective', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def interpret_old_testament_verse(request:List[Dict[str, Any]]) -> str:
    """Interpret the user-given bible verse from the Old Testament in the light of its context, together with insights of biblical Hebrew studies; an old testament bible verse / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/ot_meaning', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def expound_bible_topic(request:List[Dict[str, Any]]) -> str:
    """Expound the user-given topic in reference to the Bible; a topic must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/topic', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_theology(request:List[Dict[str, Any]]) -> str:
    """write the theological messages conveyed in the user-given content, in reference to the Bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/theology', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def study_bible_themes(request:List[Dict[str, Any]]) -> str:
    """Study Bible Themes in relation to the user content"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/themes', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_canonical_context(request:List[Dict[str, Any]]) -> str:
    """Write about canonical context of a bible book / chapter / passage; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/canon', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_related_summary(request:List[Dict[str, Any]]) -> str:
    """Write a summary on the user-given content in reference to the Bible"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/summary', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def interpret_new_testament_verse(request:List[Dict[str, Any]]) -> str:
    """Interpret the user-given bible verse from the New Testament in the light of its context, together with insights of biblical Greek studies; a new testament bible verse / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/nt_meaning', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_new_testament_highlights(request:List[Dict[str, Any]]) -> str:
    """Write Highlights in a New Testament passage in the bible; new testament bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/nt_highlights', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_applications(request:List[Dict[str, Any]]) -> str:
    """Provide detailed applications of a bible passages; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/application', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_book_introduction(request:List[Dict[str, Any]]) -> str:
    """Write a detailed introduction on a book in the bible; bible book must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/introduce_book', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_old_testament_historical_context(request:List[Dict[str, Any]]) -> str:
    """write the Bible Historical Context of a Old Testament passage in the bible; old testament bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/ot_context', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_outline(request:List[Dict[str, Any]]) -> str:
    """provide a detailed outline of a bible book / chapter / passage; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/outline', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_insights(request:List[Dict[str, Any]]) -> str:
    """Write exegetical insights in detail on a bible passage; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/insights', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.tool
def write_bible_sermon(request:List[Dict[str, Any]]) -> str:
    """Write a bible sermon based on a bible passage; bible book / chapter / passage / reference(s) must be given"""
    global agentmake, getResponse
    messages = agentmake(request, **{'instruction': 'bible/sermon', 'system': 'auto'}, **AGENTMAKE_CONFIG)
    return getResponse(messages)

@mcp.prompt
def simple_bible_study(request:str) -> PromptMessage:
    """Perform a simple bible study task"""
    global PromptMessage, TextContent
    prompt_text = f"""You are a bible study agent. You check the user request, under the `User Request` section, and resolve it with the following steps in order:
1. Call tool 'retrieve_english_bible_verses' for Bible text, 
2. Call tool 'retrieve_bible_cross_references' for Bible cross-references, 
3. Call tool 'study_old_testament_themes' for studying old testament themes or 'study_new_testament_themes' for studying new testament themes, and 
4. Call tool 'write_bible_theology' to explain its theology.

# User Request

---
{request}
---
"""
    return PromptMessage(role="user", content=TextContent(type="text", text=prompt_text))

@mcp.prompt
def bible_devotion(request:str) -> PromptMessage:
    """Write bible devotion based on user content"""
    global PromptMessage, TextContent
    prompt_text = f"""
You are a bible devotional agent. You check the user content, under the `User Content` section, and write a devotional about it with the following steps in order:

1. Analyze the themes using @study_new_testament_themes for new testament passages or @study_old_testament_themes for old testament passages.
2. Identify and explain key biblical keywords from the passage using @identify_bible_keywords.
3. Write insights for the devotion using @write_bible_insights.
4. Relate the passage to daily life using @write_bible_applications.
5. Compose a touching devotion using @write_bible_devotion.
Ensure each step is clearly addressed and the final output is cohesive and inspiring.

# User Content

---
{request}
---
"""
    return PromptMessage(role="user", content=TextContent(type="text", text=prompt_text))

mcp.run(show_banner=False)


from biblemate import config
from biblemate.uba.api import run_uba_api
from agentmake.utils.rag import get_embeddings, cosine_similarity_matrix
from typing import Union
import os, apsw, json
import numpy as np
import urllib.parse


class UBASearches:
    
    @staticmethod
    def search_data(db_file: str, sql_table: str, query: str, top_k: int=3) -> Union[list, str]:
        """search `dictionary.db` or `encyclopedia.db` for a query"""
        if not os.path.isfile(db_file):
            return "Invalid database file."

        query = urllib.parse.unquote(query).replace("「」", "/")

        keywords = {
            "dictionary.db": "DICTIONARY:::",
            "encyclopedia.db": f"ENCYCLOPEDIA:::{sql_table}:::",
            "exlb.db": f"EXLB:::{sql_table}:::",
            "collection.db": f"{'_promise' if sql_table == 'PROMISES' else '_harmony'}:::{config.default_bible}:::",
        }

        with apsw.Connection(db_file) as connection:
            cursor = connection.cursor()
            # search for an exact match
            if "+" in query:
                cmd_prefix = keywords.get(os.path.basename(db_file))
                path, _ = query.split("+", 1)
                return run_uba_api(f"{cmd_prefix}{path}")
            else:
                cursor.execute(f"SELECT * FROM {sql_table} WHERE entry = ?;", (query,))
                rows = cursor.fetchall()
            if not rows: # perform similarity search if no an exact match
                # convert query to vector
                query_vector = get_embeddings([query], config.embedding_model)[0]
                # fetch all entries
                cursor.execute(f"SELECT entry, entry_vector FROM {sql_table}")
                all_rows = cursor.fetchall()
                if not all_rows:
                    return []
                # build a matrix
                entries, entry_vectors = zip(*[(row[0], np.array(json.loads(row[1]))) for row in all_rows if row[0] and row[1]])
                document_matrix = np.vstack(entry_vectors)
                # perform a similarity search
                similarities = cosine_similarity_matrix(query_vector, document_matrix)
                top_indices = np.argsort(similarities)[::-1][:top_k]
                # return top matches
                return [entries[i] for i in top_indices]
            elif len(rows) == 1: # single exact match
                cmd_prefix = keywords.get(os.path.basename(db_file))
                path = rows[0][0]
                content = run_uba_api(f"{cmd_prefix}{path}")
                if sql_table == "exlbl" and "Click HERE for a Live Google Map" in content:
                    cursor.execute(f"SELECT lat, lng FROM {sql_table}i WHERE path = ?;", (path,))
                    lat, lng = cursor.fetchone()
                    content = content.replace("Click HERE for a Live Google Map", f"https://maps.google.com/?q={lat},{lng}&ll={lat},{lng}&z=9")
                return content
            else:
                return [f"{path}+{entry}" for path, entry, _ in rows]


from biblemate import config
from prompt_toolkit.input import create_input
from prompt_toolkit.layout import Layout, HSplit
from prompt_toolkit.widgets import Frame, Label
from prompt_toolkit.styles import Style, merge_styles
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
from prompt_toolkit.widgets import Frame, TextArea
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.layout import WindowAlign
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.markup import MarkdownLexer
from prompt_toolkit.styles import style_from_pygments_cls
from pygments.styles import get_style_by_name
from agentmake import DEFAULT_TEXT_EDITOR, edit_file, readTextFile, writeTextFile
from agentmake.etextedit import launch_async
from typing import Any, Optional
import os


async def getTextArea(input_suggestions:list=None, default_entry="", title="", multiline:bool=True, completion:Optional[Any]=None, scrollbar:bool=True, read_only:bool=False):
    """Get text area input with a border frame"""

    if config.current_prompt and not default_entry:
        default_entry = config.current_prompt
    config.current_prompt = "" # reset config.current_prompt
    completer = FuzzyCompleter(WordCompleter(input_suggestions, ignore_case=True)) if input_suggestions else None
    
    # Markdown
    pygments_style = get_style_by_name('github-dark')
    markdown_style = style_from_pygments_cls(pygments_style)
    # Define custom style
    custom_style = Style.from_dict({
        #'frame.border': '#00ff00',  # Green border
        #'frame.label': '#ffaa00 bold',  # Orange label
        #'completion-menu': 'bg:#008888 #ffffff',
        #'completion-menu.completion': 'bg:#008888 #ffffff',
        #'completion-menu.completion.current': 'bg:#00aaaa #000000',
        #"status": "reverse",
        "textarea": "bg:#1E1E1E",
    })

    style = merge_styles([markdown_style, custom_style])

    # TextArea with a completer
    text_area = TextArea(
        text=default_entry,
        style="class:textarea",
        lexer=PygmentsLexer(MarkdownLexer),
        multiline=multiline,
        scrollbar=scrollbar,
        read_only=read_only,
        completer=completer,
        complete_while_typing=config.auto_suggestions,
        focus_on_click=True,
        wrap_lines=True,
    )
    text_area.buffer.cursor_position = len(text_area.text)

    def unpack_text_chunks(_):
        openai_style = True if config.backend in ("azure", "azure_any", "custom", "deepseek", "github", "github_any", "googleai", "groq", "llamacpp", "mistral", "openai", "xai") else False
        first_event = True
        chat_response = ""
        for event in completion:
            # RETRIEVE THE TEXT FROM THE RESPONSE
            if event is None:
                continue
            elif openai_style:
                # openai
                # when open api key is invalid for some reasons, event response in string
                if isinstance(event, str):
                    answer = event
                elif hasattr(event, "data") and hasattr(event.data, "choices"): # mistralai
                    try:
                        answer = event.data.choices[0].delta.content
                    except:
                        answer = None
                elif hasattr(event, "choices") and not event.choices: # in case of the 1st event of azure's completion
                    continue
                else:
                    answer = event.choices[0].delta.content or ""
            elif hasattr(event, "type") and event.type == "content-delta" and hasattr(event, "delta"): # cohere
                answer = event.delta.message.content.text
            elif hasattr(event, "delta") and hasattr(event.delta, "text"): # anthropic
                answer = event.delta.text
            elif hasattr(event, "content_block") and hasattr(event.content_block, "text"):
                answer = event.content_block.text
            elif str(type(event)).startswith("<class 'anthropic.types"): # anthropic
                continue
            elif hasattr(event, "message"): # newer ollama python package
                answer = event.message.content
            elif isinstance(event, dict):
                if "message" in event:
                    # ollama chat
                    answer = event["message"].get("content", "")
                else:
                    # llama.cpp chat
                    answer = event["choices"][0]["delta"].get("content", "")
            elif hasattr(event, "text"):
                # vertex ai, genai
                answer = event.text
            else:
                #print(event)
                answer = None
            # STREAM THE ANSWER
            if answer is not None:
                if first_event:
                    first_event = False
                    answer = answer.lstrip()
                # update the chat response
                chat_response += answer
                # display the chunk in the text area
                text_area.buffer.insert_text(answer)
                text_area.buffer.cursor_position = len(text_area.text)

    def edit_temp_file(initial_content: str) -> str:
        config.current_prompt = ""
        temp_file = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), "temp", "edit.md")
        writeTextFile(temp_file, initial_content)
        edit_file(temp_file)
        return readTextFile(temp_file).strip()

    # Layout: include a CompletionsMenu
    root_container = HSplit(
        [
            Frame(
                text_area,
                title=title,
            ),
            Label(
                "[Ctrl+S] Send [Ctrl+Q] Exit" if title else "[Ctrl+S] Send [Ctrl+Y] Help",
                align=WindowAlign.RIGHT,
                style="fg:grey",
            ),
            CompletionsMenu(
                max_height=8,
                scroll_offset=1,
            ),
        ]
    )
    
    # Create key bindings
    bindings = KeyBindings()
    config.cursor_position = 0
    
    if not title: # for the main prompt only; these shortcuts are irrelevant for review or configuration prompts
        # help
        @bindings.add("c-y")
        def _(event):
            config.current_prompt = text_area.text
            event.app.exit(result=".help")
        # get ideas
        @bindings.add("c-g")
        def _(event):
            config.current_prompt = text_area.text
            event.app.exit(result=".ideas")
        # new chat
        @bindings.add("c-n")
        def _(event):
            config.current_prompt = text_area.text
            event.app.exit(result=".new")

    # launch editor
    @bindings.add("c-p")
    def _(event):
        config.cursor_position = text_area.buffer.cursor_position
        config.current_prompt = text_area.text
        event.app.exit(result=".editprompt")
    # exit
    @bindings.add("c-q")
    def _(event):
        event.app.exit(result=".exit")
    # submit
    @bindings.add("escape", "enter")
    @bindings.add("c-s")
    def _(event):
        if not text_area.text.strip():
            text_area.text = entry = "."
        event.app.exit(result=text_area.text.strip())
    # submit or new line
    @bindings.add("enter")
    @bindings.add("c-m")
    def _(event):
        entry = text_area.text.strip()
        if not multiline or (not title and ((entry.strip() == "." or entry.startswith(".") and entry in input_suggestions) or entry.startswith(".open ") or entry.startswith(".import "))):
            event.app.exit(result=text_area.text.strip())
        else:
            text_area.buffer.newline()
    # insert four spaces
    @bindings.add("s-tab")
    def _(event):
        buffer = event.app.current_buffer if event is not None else text_area.buffer
        buffer.insert_text("    ")
    # trigger completion
    @bindings.add("tab")
    @bindings.add("c-i")
    def _(event):
        buffer = event.app.current_buffer if event is not None else text_area.buffer
        buffer.start_completion()
    # close completion menu
    @bindings.add("escape")
    def _(event):
        buffer = event.app.current_buffer if event is not None else text_area.buffer
        buffer.cancel_completion()
    # undo
    @bindings.add("c-z")
    def _(event):
        buffer = event.app.current_buffer if event is not None else text_area.buffer
        buffer.undo()
    # reset buffer
    @bindings.add("c-r")
    def _(event):
        buffer = event.app.current_buffer if event is not None else text_area.buffer
        buffer.reset()
    # Create application
    app = Application(
        layout=Layout(root_container, focused_element=text_area),
        key_bindings=bindings,
        enable_page_navigation_bindings=True,
        style=style,
        #clipboard=PyperclipClipboard(), # not useful if mouse_support is not enabled
        #mouse_support=True, # If enabled; content outside the app becomes unscrollable
        input=create_input(always_prefer_tty=True),
        full_screen=False,
        after_render=unpack_text_chunks if completion is not None else None,
    )
    
    # Run the app
    result = await app.run_async()
    print()
    # edit in full editor
    while result == ".editprompt":
        if DEFAULT_TEXT_EDITOR == "etextedit":
            text_area.text = await launch_async(input_text=config.current_prompt, exitWithoutSaving=True, customTitle=f"BibleMate AI", startAt=config.cursor_position)
        else:
            text_area.text = edit_temp_file(config.current_prompt)
        text_area.buffer.cursor_position = len(text_area.text)
        config.current_prompt = ""
        # Run the non-full-screen text area again
        result = await app.run_async()
        print()
    # return the text content
    return result




from agentmake.utils.online import get_local_ip
import requests, os, re
from biblemate import config

DEFAULT_MODULES = {
    "bible": config.default_bible,
    "chapter": config.default_bible,
    "search": config.default_bible,
    "genesis": config.default_bible,
    "exodus": config.default_bible,
    "leviticus": config.default_bible,
    "numbers": config.default_bible,
    "deuteronomy": config.default_bible,
    "joshua": config.default_bible,
    "judges": config.default_bible,
    "ruth": config.default_bible,
    "samuel1": config.default_bible,
    "samuel2": config.default_bible,
    "kings1": config.default_bible,
    "kings2": config.default_bible,
    "chronicles1": config.default_bible,
    "chronicles2": config.default_bible,
    "ezra": config.default_bible,
    "nehemiah": config.default_bible,
    "esther": config.default_bible,
    "job": config.default_bible,
    "psalms": config.default_bible,
    "proverbs": config.default_bible,
    "ecclesiastes": config.default_bible,
    "songs": config.default_bible,
    "isaiah": config.default_bible,
    "jeremiah": config.default_bible,
    "lamentations": config.default_bible,
    "ezekiel": config.default_bible,
    "daniel": config.default_bible,
    "hosea": config.default_bible,
    "joel": config.default_bible,
    "amos": config.default_bible,
    "obadiah": config.default_bible,
    "jonah": config.default_bible,
    "micah": config.default_bible,
    "nahum": config.default_bible,
    "habakkuk": config.default_bible,
    "zephaniah": config.default_bible,
    "haggai": config.default_bible,
    "zechariah": config.default_bible,
    "malachi": config.default_bible,
    "matthew": config.default_bible,
    "mark": config.default_bible,
    "luke": config.default_bible,
    "john": config.default_bible,
    "acts": config.default_bible,
    "romans": config.default_bible,
    "corinthians1": config.default_bible,
    "corinthians2": config.default_bible,
    "galatians": config.default_bible,
    "ephesians": config.default_bible,
    "philippians": config.default_bible,
    "colossians": config.default_bible,
    "thessalonians1": config.default_bible,
    "thessalonians2": config.default_bible,
    "timothy1": config.default_bible,
    "timothy2": config.default_bible,
    "titus": config.default_bible,
    "philemon": config.default_bible,
    "hebrews": config.default_bible,
    "james": config.default_bible,
    "peter1": config.default_bible,
    "peter2": config.default_bible,
    "john1": config.default_bible,
    "john2": config.default_bible,
    "john3": config.default_bible,
    "jude": config.default_bible,
    "revelation": config.default_bible,
    "commentary": config.default_commentary,
    "encyclopedia": config.default_encyclopedia,
    "lexicon": config.default_lexicon,
}

# api
def run_uba_api(command: str, html=False) -> str:
    UBA_API_LOCAL_PORT = int(os.getenv("UBA_API_LOCAL_PORT")) if os.getenv("UBA_API_LOCAL_PORT") else 8080
    UBA_API_ENDPOINT = os.getenv("UBA_API_ENDPOINT") if os.getenv("UBA_API_ENDPOINT") else f"http://{get_local_ip()}:{UBA_API_LOCAL_PORT}/plain" # use dynamic local ip if endpoint is not specified
    UBA_API_TIMEOUT = int(os.getenv("UBA_API_TIMEOUT")) if os.getenv("UBA_API_TIMEOUT") else 10
    UBA_API_PRIVATE_KEY = os.getenv("UBA_API_PRIVATE_KEY") if os.getenv("UBA_API_PRIVATE_KEY") else ""

    endpoint = UBA_API_ENDPOINT
    if html:
        endpoint = endpoint.replace("/plain", "/html")
    private = f"private={UBA_API_PRIVATE_KEY}&" if UBA_API_PRIVATE_KEY else ""
    url = f"""{endpoint}?{private}cmd={command}"""
    try:
        response = requests.get(url, timeout=UBA_API_TIMEOUT)
        response.encoding = "utf-8"
        content = response.text.strip()
        if command.lower().startswith("data:::"):
            return content.replace("\n", "\n- ")
        content = re.sub(r"\n([0-9]+?) \(([^\(\)]+?)\)", r"\n- `\1` (`\2`)", content)
        content = re.sub(r"^([0-9]+?) \(([^\(\)]+?)\)", r"- `\1` (`\2`)", content)
        content = re.sub(r"\n\(([^\(\)]+?)\)", r"\n- (`\1`)", content)
        return re.sub(r"^\(([^\(\)]+?)\)", r"- (`\1`)", content)
    except Exception as err:
        return f"An error occurred: {err}"